==> jvterm-0.2.5.tar.gz: jvterm-0.2.5/console.c <==
/* console.c - implement a virtual console
 *
 * Copyright (C) 2003, 2005  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 *
 * $Id: console.c 6725 2005-11-09 23:15:58Z voss $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>		/* for BEL below (remove eventually) */
#include <stdint.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>

#include "base.h"
#include "backend.h"
#include "ui.h"

#define JV_ATTR_STANDOUT JV_ATTR_REVERSE

#define NPAR 16
struct console_data {
  struct shell_client_data *scd;

  struct buffer *b;
  struct converter *conv;

  enum { cs_normal, cs_esc, cs_getpars,
	 cs_bracket, cs_tekcarb, cs_image, cs_imgname, cs_imgzoom } state;
  
  Uint8  attr;			/* current attributes */
  Uint8  xattr;			/* current overlay attributes */
  Uint8  colors;		/* current forground and background color */
  
  unsigned  par[NPAR];
  int  npar;

  int  cx_saved, cy_saved;	/* saved cursor position */

  char *fname;
  size_t  fname_used, fname_allocated;
};


static void
console_reset (struct console_data *cd)
/* Reset the console and all subsystems to their initial states.  */
{
  buffer_reset (cd->b);
  converter_reset (cd->conv);

  cd->state = cs_normal;
  
  cd->attr = 0;
  cd->xattr = 0;
  cd->colors = DEFAULT_COLOR;
  
  cd->npar = 0;

  cd->cx_saved = cd->cy_saved = -1;

  cd->fname = NULL;
  cd->fname_allocated = cd->fname_used = 0;
}

static void
csi_m (struct console_data *cd, int code)
/* Compatibility code for broken applications.  We support the
 * "ECMA-48 SGR" codes as documented in the console_codes(4) manual
 * page, but we do not advertise the corresponding codes via
 * terminfo.  */
{
  switch (code) {
  case 0:
    cd->attr = 0;
    cd->colors = DEFAULT_COLOR;
    break;
  case 1:
    cd->attr |= JV_ATTR_BRIGHT;
    break;
  case 2:
    cd->attr |= JV_ATTR_DIM;
    break;
  case 4:
    cd->attr |= JV_ATTR_UNDERLINE;
    break;
  case 5:
    cd->attr |= JV_ATTR_BLINK;
    break;
  case 7:
    cd->attr |= JV_ATTR_REVERSE;
    break;
  case 10:
  case 11:
  case 12:
    break;
  case 21:
  case 22:
    cd->attr &= ~(JV_ATTR_BRIGHT|JV_ATTR_DIM);
    break;
  case 24:
    cd->attr &= ~JV_ATTR_UNDERLINE;
    break;
  case 25:
    cd->attr &= ~JV_ATTR_BLINK;
    break;
  case 27:
    cd->attr &= ~JV_ATTR_REVERSE;
    break;
  case 30:
    cd->colors = JV_COLOR_PAIR (JV_COLOR_BLACK,
				GET_COLOR_BG(cd->colors));
    break;
  case 31:
    cd->colors = JV_COLOR_PAIR (JV_COLOR_RED,
				GET_COLOR_BG(cd->colors));
    break;
  case 32:
    cd->colors = JV_COLOR_PAIR (JV_COLOR_GREEN,
				GET_COLOR_BG(cd->colors));
    break;
  case 33:
    cd->colors = JV_COLOR_PAIR (JV_COLOR_YELLOW,
				GET_COLOR_BG(cd->colors));
    break;
  case 34:
    cd->colors = JV_COLOR_PAIR (JV_COLOR_BLUE,
				GET_COLOR_BG(cd->colors));
    break;
  case 35:
    cd->colors = JV_COLOR_PAIR (JV_COLOR_MAGENTA,
				GET_COLOR_BG(cd->colors));
    break;
  case 36:
    cd->colors = JV_COLOR_PAIR (JV_COLOR_CYAN,
				GET_COLOR_BG(cd->colors));
    break;
  case 37:
    cd->colors = JV_COLOR_PAIR (JV_COLOR_WHITE,
				GET_COLOR_BG(cd->colors));
    break;
  case 38:
    cd->attr |= JV_ATTR_UNDERLINE;
    cd->colors = JV_COLOR_PAIR (JV_COLOR_WHITE,
				GET_COLOR_BG(cd->colors));
    break;
  case 39:
    cd->attr &= ~JV_ATTR_UNDERLINE;
    cd->colors = JV_COLOR_PAIR (JV_COLOR_WHITE,
				GET_COLOR_BG(cd->colors));
    break;
  case 40:
  case 49:
    cd->colors = JV_COLOR_PAIR (GET_COLOR_BG(cd->colors),
				JV_COLOR_BLACK);
    break;
  case 41:
    cd->colors = JV_COLOR_PAIR (GET_COLOR_BG(cd->colors),
				JV_COLOR_RED);
    break;
  case 42:
    cd->colors = JV_COLOR_PAIR (GET_COLOR_BG(cd->colors),
				JV_COLOR_GREEN);
    break;
  case 43:
    cd->colors = JV_COLOR_PAIR (GET_COLOR_BG(cd->colors),
				JV_COLOR_YELLOW);
    break;
  case 44:
    cd->colors = JV_COLOR_PAIR (GET_COLOR_BG(cd->colors),
				JV_COLOR_BLUE);
    break;
  case 45:
    cd->colors = JV_COLOR_PAIR (GET_COLOR_BG(cd->colors),
				JV_COLOR_MAGENTA);
    break;
  case 46:
    cd->colors = JV_COLOR_PAIR (GET_COLOR_BG(cd->colors),
				JV_COLOR_CYAN);
    break;
  case 47:
    cd->colors = JV_COLOR_PAIR (GET_COLOR_BG(cd->colors),
				JV_COLOR_WHITE);
    break;
  }
}

static void
cooked_putchar (struct console_data *cd, Uint16 c)
/* Decode the control sequences.  This digests the output character C
 * (in UCS-2 encoding), updates the state in CD, and emits the
 * character if it is not part of a control sequence.  */
{
  switch (cd->state) {
  case cs_normal:
    if (c >= 32) {		/* normal character output */
      struct charcell  cc;

      if (c == 0x200B || c == 0xFEFF) {
	/* ignore zero width characters */
	buffer_clear_basechar (cd->b);
	break;
      }

      if (c == 0x200C || c == 0x200D) {
	/* ignore ZERO WIDTH JOINER and ZERO WIDTH NON-JOINER */
	break;
      }

      cc.c = c;
      cc.attr = cd->attr|cd->xattr;
      cc.colors = cd->colors;
      buffer_putc (cd->b, cc);
    } else {			/* control sequences */
      buffer_clear_basechar (cd->b);
      switch (c) {
      case 0:			/* NUL */
	break;
      case 5:			/* ENQ */
	shell_send_char (cd->scd, 6); /* send ACK */
	break;
      case 7:			/* BEL */
	putchar ('\a');		/* TODO */
	fflush (stdout);
	break;
      case 8:			/* BS */
	if (cd->b->cx > 0) {
	  buffer_set_cursor (cd->b, cd->b->cx-1, cd->b->cy);
	} else if (cd->b->cy > 0 && cd->b->am_flag) {
	  buffer_set_cursor (cd->b, cd->b->cols-1, cd->b->cy-1);
	}
	break;
      case 9:			/* TAB */
	{
	  int  x;
	  x = ((cd->b->cx/8)+1)*8;
	  if (x > cd->b->cols-1)  x = cd->b->cols-1;
	  buffer_set_cursor (cd->b, x, cd->b->cy);
	}
	break;
      case 10:			/* LF */
      case 11:			/* VT */
	buffer_lf (cd->b);
	break;
      case 12:			/* FF */
	buffer_crlf (cd->b);
	break;
      case 13:			/* CR */
	buffer_set_cursor (cd->b, 0, cd->b->cy);
	break;
      case 27:			/* ESC */
	cd->state = cs_esc;
	break;
      default:
	warning ("received invalid control code '^%c'", c+'A'-1);
	break;
      }
    }
    break;

  case cs_esc:
    cd->state = cs_normal;
    switch (c) {
    case '[':
      cd->npar = 0;
      cd->par[0] = 0;
      cd->state = cs_getpars;
      break;
    case '(':
      cd->state = cs_bracket;
      break;
    case ')':
      cd->state = cs_tekcarb;
      break;
    case '!':
      cd->fname_used = 0;
      cd->fname_allocated = 256;
      cd->fname = xmalloc(cd->fname_allocated);
      cd->state = cs_image;
      break;

#include "codes1.h"

    default:
      if (32<c && c<127) {
	warning ("received invalid sequence 'ESC %c'", c);
      } else {
	warning ("received invalid sequence 'ESC u%04x'", c);
      }
      break;
    }
    break;

  case cs_getpars:
    if (c>='0' && c<='9') {
      cd->par[cd->npar] *= 10;
      cd->par[cd->npar] += c-'0';
      break;
    }
    cd->npar++;
    cd->par[cd->npar] = 0;
    if (c==';' && cd->npar<NPAR-1)  break;

    switch (c) {
#include "codes2.h"

    case 'm':
      /* Emulate the Linux console driver's "CSI m" command.
       * Some broken programs need this.  */
      {
	int i;
	for (i=0; i<cd->npar; ++i) csi_m (cd, cd->par[i]);
      }
      break;
      
    default:
      if (32<c && c<127) {
	warning ("received invalid sequence 'ESC [ %c'", c);
      } else {
	warning ("received invalid sequence 'ESC [ u%04x'", c);
      }
      break;
    }
    cd->state = cs_normal;
    break;

  case cs_bracket:
    if (32<c && c<127) {
      warning ("received invalid sequence 'ESC ( %c'", c);
    } else {
      warning ("received invalid sequence 'ESC ( u%04x'", c);
    }
    cd->state = cs_normal;
    break;

  case cs_tekcarb:
    if (c != '0') {
      if (32<c && c<127) {
	warning ("received invalid sequence 'ESC ) %c'", c);
      } else {
	warning ("received invalid sequence 'ESC ) u%04x'", c);
      }
    }
    cd->state = cs_normal;
    break;

  case cs_image:
    if (c == 'f') {
      cd->state = cs_imgname;
    } else if (c == 'z') {
      cd->par[0] = 0;
      cd->state = cs_imgzoom;
    } else if (c == 'q') {
      buffer_image_load (cd->b, NULL);
      cd->state = cs_normal;
    } else if (c == 'u') {
      buffer_image_scroll (cd->b, 0, -32);
      cd->state = cs_normal;
    } else if (c == 'd') {
      buffer_image_scroll (cd->b, 0, +32);
      cd->state = cs_normal;
    } else if (c == 'l') {
      buffer_image_scroll (cd->b, -32, 0);
      cd->state = cs_normal;
    } else if (c == 'r') {
      buffer_image_scroll (cd->b, +32, 0);
      cd->state = cs_normal;
    } else {
      if (32<c && c<127) {
	warning ("received invalid sequence 'ESC ! %c'", c);
      } else {
	warning ("received invalid sequence 'ESC ! u%04x'", c);
      }
      cd->state = cs_normal;
    }
    break;
    
  case cs_imgname:
    if (cd->fname_used >= cd->fname_allocated) {
      cd->fname_allocated *= 2;
      cd->fname = xrealloc (cd->fname, cd->fname_allocated);
    }
    if (c == '\n' || c == '\r') {
      cd->fname[cd->fname_used++] = '\0';
      buffer_image_load (cd->b, cd->fname);
    }
    if (c<32 || c>=127) {
      if (c != '\n' && c != '\r') {
	message ("image file name parsing aborted on char u%04x", c);
      }
      xfree (cd->fname);
      cd->fname = NULL;
      cd->fname_allocated = cd->fname_used = 0;
      cd->state = cs_normal;
    } else {
      cd->fname[cd->fname_used++] = c;
    }
    break;

  case cs_imgzoom:
    if (c>='0' && c<='9') {
      cd->par[0] *= 10;
      cd->par[0] += c-'0';
      break;
    }
    buffer_image_zoom (cd->b, 0.01*cd->par[0]);
    cd->state = cs_normal;
    break;
    
  default:
    cd->state = cs_normal;
    break;
  }
}

static const uint32_t unicode_replacement_character = 0xFFFD;

static ssize_t
emit_ucs4 (const void *buffer, size_t size, void *client_data)
{
  struct console_data *cd = client_data;
  const uint32_t *output = buffer;
  int  i;

  assert (size % sizeof(uint32_t) == 0);

  for (i=0; i<size/sizeof(uint32_t); ++i) {
    uint32_t c = output[i];
    cooked_putchar (cd, c>0xFFFF ? unicode_replacement_character : c);
  }

  return  size;
}

static int
is_bigendian (void)
{
  union {
    unsigned char as_char[2];
    short as_short;
  } test;
  test.as_short = 0x1234;
  return  (test.as_char[0] != 0x34);
}

struct console_data *
console_open (struct buffer *buf, struct shell_client_data *scd,
	      const char *conv_id)
{
  struct console_data *cd;
  const char *ucs4;

  cd = xnew (struct console_data);
  cd->scd = scd;
  cd->b = buf;

  ucs4 = is_bigendian() ? "UCS-4BE" : "UCS-4LE";
  cd->conv = open_converter (conv_id,
			     scd->child_charset, ucs4,
			     &unicode_replacement_character, 4,
			     emit_ucs4, cd);
  if (! cd->conv) {
    fatal ("cannot get converter from %s to %s", scd->child_charset, ucs4);
  }
  console_reset (cd);
  
  return  cd;
}

int
console_close (struct console_data *cd)
{
  close_converter (cd->conv);
  xfree (cd);

  return  0;
}

ssize_t
console_copy (struct console_data *cd, int fd)
/* Send data to the console.
 * This function does the output character set conversion
 * and then implicitly calls 'cooked_putchar' on the result.  */
{
  return  converter_copy (cd->conv, fd);
}

==> jvterm-0.2.6.tar.gz: jvterm-0.2.6/console.c <==
/* console.c - implement a virtual console
 *
 * Copyright (C) 2003, 2005  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 *
 * $Id: console.c 6731 2005-11-10 20:28:57Z voss $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>		/* for BEL below (remove eventually) */
#include <stdint.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>

#include "base.h"
#include "backend.h"
#include "ui.h"

#define JV_ATTR_STANDOUT JV_ATTR_REVERSE

#define NPAR 16
struct console_data {
  struct shell_client_data *scd;

  struct buffer *b;
  struct converter *conv;

  enum { cs_normal, cs_esc, cs_getpars,
	 cs_bracket, cs_tekcarb, cs_image, cs_imgname, cs_imgzoom } state;
  
  Uint8  attr;			/* current attributes */
  Uint8  xattr;			/* current overlay attributes */
  Uint8  colors;		/* current forground and background color */
  
  unsigned  par[NPAR];
  int  npar;

  int  cx_saved, cy_saved;	/* saved cursor position */

  char *fname;
  size_t  fname_used, fname_allocated;
};


static void
console_reset (struct console_data *cd)
/* Reset the console and all subsystems to their initial states.  */
{
  buffer_reset (cd->b);
  converter_reset (cd->conv);

  cd->state = cs_normal;
  
  cd->attr = 0;
  cd->xattr = 0;
  cd->colors = DEFAULT_COLOR;
  
  cd->npar = 0;

  cd->cx_saved = cd->cy_saved = -1;

  cd->fname = NULL;
  cd->fname_allocated = cd->fname_used = 0;
}

static void
csi_m (struct console_data *cd, int code)
/* Compatibility code for broken applications.  We support the
 * "ECMA-48 SGR" codes as documented in the console_codes(4) manual
 * page, but we do not advertise the corresponding codes via
 * terminfo.  */
{
  switch (code) {
  case 0:
    cd->attr = 0;
    cd->colors = DEFAULT_COLOR;
    break;
  case 1:
    cd->attr |= JV_ATTR_BRIGHT;
    break;
  case 2:
    cd->attr |= JV_ATTR_DIM;
    break;
  case 4:
    cd->attr |= JV_ATTR_UNDERLINE;
    break;
  case 5:
    cd->attr |= JV_ATTR_BLINK;
    break;
  case 7:
    cd->attr |= JV_ATTR_REVERSE;
    break;
  case 10:
  case 11:
  case 12:
    break;
  case 21:
  case 22:
    cd->attr &= ~(JV_ATTR_BRIGHT|JV_ATTR_DIM);
    break;
  case 24:
    cd->attr &= ~JV_ATTR_UNDERLINE;
    break;
  case 25:
    cd->attr &= ~JV_ATTR_BLINK;
    break;
  case 27:
    cd->attr &= ~JV_ATTR_REVERSE;
    break;
  case 30:
    cd->colors = JV_COLOR_PAIR (JV_COLOR_BLACK,
				GET_COLOR_BG(cd->colors));
    break;
  case 31:
    cd->colors = JV_COLOR_PAIR (JV_COLOR_RED,
				GET_COLOR_BG(cd->colors));
    break;
  case 32:
    cd->colors = JV_COLOR_PAIR (JV_COLOR_GREEN,
				GET_COLOR_BG(cd->colors));
    break;
  case 33:
    cd->colors = JV_COLOR_PAIR (JV_COLOR_YELLOW,
				GET_COLOR_BG(cd->colors));
    break;
  case 34:
    cd->colors = JV_COLOR_PAIR (JV_COLOR_BLUE,
				GET_COLOR_BG(cd->colors));
    break;
  case 35:
    cd->colors = JV_COLOR_PAIR (JV_COLOR_MAGENTA,
				GET_COLOR_BG(cd->colors));
    break;
  case 36:
    cd->colors = JV_COLOR_PAIR (JV_COLOR_CYAN,
				GET_COLOR_BG(cd->colors));
    break;
  case 37:
    cd->colors = JV_COLOR_PAIR (JV_COLOR_WHITE,
				GET_COLOR_BG(cd->colors));
    break;
  case 38:
    cd->attr |= JV_ATTR_UNDERLINE;
    cd->colors = JV_COLOR_PAIR (JV_COLOR_WHITE,
				GET_COLOR_BG(cd->colors));
    break;
  case 39:
    cd->attr &= ~JV_ATTR_UNDERLINE;
    cd->colors = JV_COLOR_PAIR (JV_COLOR_WHITE,
				GET_COLOR_BG(cd->colors));
    break;
  case 40:
  case 49:
    cd->colors = JV_COLOR_PAIR (GET_COLOR_BG(cd->colors),
				JV_COLOR_BLACK);
    break;
  case 41:
    cd->colors = JV_COLOR_PAIR (GET_COLOR_BG(cd->colors),
				JV_COLOR_RED);
    break;
  case 42:
    cd->colors = JV_COLOR_PAIR (GET_COLOR_BG(cd->colors),
				JV_COLOR_GREEN);
    break;
  case 43:
    cd->colors = JV_COLOR_PAIR (GET_COLOR_BG(cd->colors),
				JV_COLOR_YELLOW);
    break;
  case 44:
    cd->colors = JV_COLOR_PAIR (GET_COLOR_BG(cd->colors),
				JV_COLOR_BLUE);
    break;
  case 45:
    cd->colors = JV_COLOR_PAIR (GET_COLOR_BG(cd->colors),
				JV_COLOR_MAGENTA);
    break;
  case 46:
    cd->colors = JV_COLOR_PAIR (GET_COLOR_BG(cd->colors),
				JV_COLOR_CYAN);
    break;
  case 47:
    cd->colors = JV_COLOR_PAIR (GET_COLOR_BG(cd->colors),
				JV_COLOR_WHITE);
    break;
  }
}

static void
cooked_putchar (struct console_data *cd, Uint16 c)
/* Decode the control sequences.  This digests the output character C
 * (in UCS-2 encoding), updates the state in CD, and emits the
 * character if it is not part of a control sequence.  */
{
  switch (cd->state) {
  case cs_normal:
    if (c >= 32) {		/* normal character output */
      struct charcell  cc;

      if (c == 0x200B || c == 0xFEFF) {
	/* ignore zero width characters */
	buffer_clear_basechar (cd->b);
	break;
      }

      if (c == 0x200C || c == 0x200D) {
	/* ignore ZERO WIDTH JOINER and ZERO WIDTH NON-JOINER */
	break;
      }

      cc.c = c;
      cc.attr = cd->attr|cd->xattr;
      cc.colors = cd->colors;
      buffer_putc (cd->b, cc);
    } else {			/* control sequences */
      buffer_clear_basechar (cd->b);
      switch (c) {
      case 0:			/* NUL */
	break;
      case 5:			/* ENQ */
	shell_send_char (cd->scd, 6); /* send ACK */
	break;
      case 7:			/* BEL */
	putchar ('\a');		/* TODO */
	fflush (stdout);
	break;
      case 8:			/* BS */
	if (cd->b->cx > 0) {
	  buffer_set_cursor (cd->b, cd->b->cx-1, cd->b->cy);
	} else if (cd->b->cy > 0 && cd->b->am_flag) {
	  buffer_set_cursor (cd->b, cd->b->cols-1, cd->b->cy-1);
	}
	break;
      case 9:			/* TAB */
	{
	  int  x;
	  x = ((cd->b->cx/8)+1)*8;
	  if (x > cd->b->cols-1)  x = cd->b->cols-1;
	  buffer_set_cursor (cd->b, x, cd->b->cy);
	}
	break;
      case 10:			/* LF */
      case 11:			/* VT */
	buffer_lf (cd->b);
	break;
      case 12:			/* FF */
	buffer_crlf (cd->b);
	break;
      case 13:			/* CR */
	buffer_set_cursor (cd->b, 0, cd->b->cy);
	break;
      case 27:			/* ESC */
	cd->state = cs_esc;
	break;
      default:
	warning ("received invalid control code '^%c'", c+'A'-1);
	break;
      }
    }
    break;

  case cs_esc:
    cd->state = cs_normal;
    switch (c) {
    case '[':
      cd->npar = 0;
      cd->par[0] = 0;
      cd->state = cs_getpars;
      break;
    case '(':
      cd->state = cs_bracket;
      break;
    case ')':
      cd->state = cs_tekcarb;
      break;
    case '!':
      cd->fname_used = 0;
      cd->fname_allocated = 256;
      cd->fname = xmalloc(cd->fname_allocated);
      cd->state = cs_image;
      break;

#include "codes1.h"

    default:
      if (32<c && c<127) {
	warning ("received invalid sequence 'ESC %c'", c);
      } else {
	warning ("received invalid sequence 'ESC u%04x'", c);
      }
      break;
    }
    break;

  case cs_getpars:
    if (c>='0' && c<='9') {
      cd->par[cd->npar] *= 10;
      cd->par[cd->npar] += c-'0';
      break;
    }
    cd->npar++;
    cd->par[cd->npar] = 0;
    if (c==';' && cd->npar<NPAR-1)  break;

    switch (c) {
#include "codes2.h"

    case 'm':
      /* Emulate the Linux console driver's "CSI m" command.
       * Some broken programs need this.  */
      {
	int i;
	for (i=0; i<cd->npar; ++i) csi_m (cd, cd->par[i]);
      }
      break;
      
    default:
      if (32<c && c<127) {
	warning ("received invalid sequence 'ESC [ %c'", c);
      } else {
	warning ("received invalid sequence 'ESC [ u%04x'", c);
      }
      break;
    }
    cd->state = cs_normal;
    break;

  case cs_bracket:
    if (32<c && c<127) {
      warning ("received invalid sequence 'ESC ( %c'", c);
    } else {
      warning ("received invalid sequence 'ESC ( u%04x'", c);
    }
    cd->state = cs_normal;
    break;

  case cs_tekcarb:
    if (c != '0') {
      if (32<c && c<127) {
	warning ("received invalid sequence 'ESC ) %c'", c);
      } else {
	warning ("received invalid sequence 'ESC ) u%04x'", c);
      }
    }
    cd->state = cs_normal;
    break;

  case cs_image:
    if (c == 'f') {
      cd->state = cs_imgname;
    } else if (c == 'z') {
      cd->par[0] = 0;
      cd->state = cs_imgzoom;
    } else if (c == 'q') {
      buffer_image_load (cd->b, NULL);
      cd->state = cs_normal;
    } else if (c == 'u') {
      buffer_image_scroll (cd->b, 0, -32);
      cd->state = cs_normal;
    } else if (c == 'd') {
      buffer_image_scroll (cd->b, 0, +32);
      cd->state = cs_normal;
    } else if (c == 'l') {
      buffer_image_scroll (cd->b, -32, 0);
      cd->state = cs_normal;
    } else if (c == 'r') {
      buffer_image_scroll (cd->b, +32, 0);
      cd->state = cs_normal;
    } else {
      if (32<c && c<127) {
	warning ("received invalid sequence 'ESC ! %c'", c);
      } else {
	warning ("received invalid sequence 'ESC ! u%04x'", c);
      }
      cd->state = cs_normal;
    }
    break;
    
  case cs_imgname:
    if (cd->fname_used >= cd->fname_allocated) {
      cd->fname_allocated *= 2;
      cd->fname = xrealloc (cd->fname, cd->fname_allocated);
    }
    if (c == '\n' || c == '\r') {
      cd->fname[cd->fname_used++] = '\0';
      buffer_image_load (cd->b, cd->fname);
    }
    if (c<32 || c>=127) {
      if (c != '\n' && c != '\r') {
	message ("image file name parsing aborted on char u%04x", c);
      }
      xfree (cd->fname);
      cd->fname = NULL;
      cd->fname_allocated = cd->fname_used = 0;
      cd->state = cs_normal;
    } else {
      cd->fname[cd->fname_used++] = c;
    }
    break;

  case cs_imgzoom:
    if (c>='0' && c<='9') {
      cd->par[0] *= 10;
      cd->par[0] += c-'0';
      break;
    }
    buffer_image_zoom (cd->b, 0.01*cd->par[0]);
    cd->state = cs_normal;
    break;
    
  default:
    cd->state = cs_normal;
    break;
  }
}

static const uint32_t unicode_replacement_character = 0xFFFD;

static ssize_t
emit_ucs4 (const void *buffer, size_t size, void *client_data)
{
  struct console_data *cd = client_data;
  const uint32_t *output = buffer;
  int  i;

  assert (size % sizeof(uint32_t) == 0);

  for (i=0; i<size/sizeof(uint32_t); ++i) {
    uint32_t c = output[i];
    cooked_putchar (cd, c>0xFFFF ? unicode_replacement_character : c);
  }

  return  size;
}

static int
is_bigendian (void)
{
  union {
    unsigned char as_char[2];
    short as_short;
  } test;
  test.as_short = 0x1234;
  return  (test.as_char[0] != 0x34);
}

struct console_data *
console_open (struct buffer *buf, struct shell_client_data *scd,
	      const char *conv_id)
{
  struct console_data *cd;
  const char *ucs4;

  cd = xnew (struct console_data);
  cd->scd = scd;
  cd->b = buf;

  ucs4 = is_bigendian() ? "UCS-4BE" : "UCS-4LE";
  cd->conv = open_converter (conv_id,
			     scd->child_charset, ucs4,
			     &unicode_replacement_character, 4,
			     emit_ucs4, cd);
  if (! cd->conv) {
    fatal ("cannot get converter from %s to %s", scd->child_charset, ucs4);
  }
  console_reset (cd);
  
  return  cd;
}

int
console_close (struct console_data *cd)
{
  close_converter (cd->conv);
  xfree (cd);

  return  0;
}

ssize_t
console_copy (struct console_data *cd, int fd)
/* Send data to the console.
 * This function does the output character set conversion
 * and then implicitly calls 'cooked_putchar' on the result.  */
{
  ssize_t  rc;

  rc = converter_copy (cd->conv, fd);
  if (rc == (ssize_t)-1 && errno != EIO) {
    error ("console_copy (vt%d): copy failed: %m", cd->scd->vt+1);
  }
  
  return  rc;
}
